package no.jabber.eclipseplugin.linking;

import java.io.File;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Hashtable;
import java.util.Iterator;

import no.jabber.eclipseplugin.GuiUtil;
import no.jabber.eclipseplugin.projectmanagement.Project;
import no.jabber.eclipseplugin.projectmanagement.ProjectConfig;

import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.runtime.IPath;
import org.eclipse.core.runtime.Path;
import org.eclipse.jdt.core.ICompilationUnit;
import org.eclipse.jdt.core.IField;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IMethod;
import org.eclipse.jdt.core.JavaModelException;

/**
 * A hyperlink of this type can point to a resource in a project.
 */
public class EclipseLink 
{
	private String strLink;
	private URI uri; 
	private String strQry; 
	private ArrayList<LinkParameter> alLinkParams = new ArrayList<LinkParameter>();
	private Hashtable<String,LinkParameter> htLinkParamsByParamType = new Hashtable<String,LinkParameter>();
	
	public static String ECLIPSE_SCHEME = "eclipse";
	public static String ECLIPSE_WKSPC = "workspace"; // To indicate that the link links to a workspace. Probably no other option here. Possibly for future extensions.

	public EclipseLink(String strPath, ArrayList<LinkParameter> params, Project pr) throws URISyntaxException
	{
		if((params != null) && (params.size() > 0) )
		{
			this.alLinkParams = params;
			mapParametersToHt();
			this.strQry = makeQryString(params);
		}
		if(pr != null) 
		{
			// Replace local project name with common name from the Project object
			int nFirstSegmentStart = strPath.indexOf("/");
			int nFirstSegmentEnd = strPath.indexOf("/", nFirstSegmentStart+1);
			strPath = "/" +pr.getCommonName() + strPath.substring(nFirstSegmentEnd);
			//pr.getCommonName();
		}
		
		this.uri = new URI(EclipseLink.ECLIPSE_SCHEME,EclipseLink.ECLIPSE_WKSPC,strPath,this.strQry,null);
		this.strLink = this.uri.toString();
	}
	
	public EclipseLink(String strLink) throws URISyntaxException
	{
		ConstructLink(strLink);
	}
	
	public String makeQryString(ArrayList<LinkParameter> params)
	{
		StringBuilder sb = new StringBuilder();
		for (Iterator iter = params.iterator(); iter.hasNext();)
		{
			LinkParameter p = (LinkParameter) iter.next();
			sb.append(p.getParamName());
			sb.append("=");
			sb.append(p.getParamValue());
			sb.append("&");
		}
		return sb.toString().substring(0,sb.length()-1);
	}
	
	private void ConstructLink(String strLink) throws URISyntaxException
	{
		this.strLink = strLink;
		this.uri = new URI(this.strLink);
		this.strQry = this.uri.getQuery();
		if(this.strQry != null)
		{
			// Make parameter objects
			String[] aParams = this.strQry.split("&");
			for (int i = 0; i < aParams.length; i++) 
			{
				String[] param = aParams[i].split("=");
				String strParam = param[0];
				String strValue = param[1];
				if(strValue == null) strValue = "";
				alLinkParams.add(new LinkParameter(strParam,strValue));
			}
			mapParametersToHt();
		}
		else alLinkParams = null;
	}

	/**
	 * Maps the collection of parameters to a hashtable.
	 */
	private void mapParametersToHt()
	{
		for (int i = 0; i < alLinkParams.size(); i++)
		{
			LinkParameter param = alLinkParams.get(i);
			String strPName = param.getParamName();
			if		(strPName.equalsIgnoreCase(IEclipseLinkParamsTypes.FILE_TYPE)) htLinkParamsByParamType.put(IEclipseLinkParamsTypes.FILE_TYPE,param);
			else if (strPName.equalsIgnoreCase(IEclipseLinkParamsTypes.LINE_NUMBER)) htLinkParamsByParamType.put(IEclipseLinkParamsTypes.LINE_NUMBER,param);
			else if (strPName.equalsIgnoreCase(IEclipseLinkParamsTypes.METHOD_NAME)) htLinkParamsByParamType.put(IEclipseLinkParamsTypes.METHOD_NAME,param);
			else if (strPName.equalsIgnoreCase(IEclipseLinkParamsTypes.FIELD_NAME)) htLinkParamsByParamType.put(IEclipseLinkParamsTypes.FIELD_NAME,param);
		}
	}
	
	/**
	 * The link as a string
	 */
	public String getLinkAsString()
	{
		return this.strLink;
	}
	
	/**
	 * This method only checks if the link is on the format "eclipse://workspace/...", NOT if the resource actually exist
	 * @return
	 */
	public boolean linksToWorkspaceResource()
	{
		return this.uri.getHost().equalsIgnoreCase(EclipseLink.ECLIPSE_WKSPC);
	}
	
	public ArrayList<LinkParameter> getParameters()
	{
		return this.alLinkParams;
	}
	
	public LinkParameter getParameter(String strParamType)
	{
		return htLinkParamsByParamType.get(strParamType);
	}
	
	public String getQry()
	{
		return this.strQry;
	}
	
	public String getPath()
	{
		return uri.getPath();
	}
	
	/**
	 * Gets the file type for the file which the link points to
	 */
	public String getFileType()
	{
		String strFileType = IFileTypes.FT_UNKNOWN; 
		LinkParameter lpFileType = this.getParameter(IEclipseLinkParamsTypes.FILE_TYPE);
		if(lpFileType != null) strFileType = lpFileType.getParamValue();
		else
		{
			// Detect file type based on extension
			Path path = new Path(this.uri.getPath());
			strFileType = path.getFileExtension();
		}
		return strFileType;
	}
	
	/**
	 * Checks a string if it is an eclipse link
	 * @param strUri
	 * @return
	 */
	public static boolean isEclipseLink(String strUri)
	{
		URI uri = null;
		try 
		{
			uri = new URI(strUri);
		} 
		catch (URISyntaxException e) 
		{
			return false;
		}
		if(uri.getScheme().equals(EclipseLink.ECLIPSE_SCHEME)) return true;
		else return false;
	}

	public IResource getResource()
	{
		if(resourceExists()) return res;
		return null;
	}
	private transient IResource res;
	
	private boolean checkExistsOnDisk(String strRelativePath)
	{
		Path pth = new Path(strRelativePath);
		// Check if file really exists on disk
		Path pthTest = new Path(GuiUtil.getCurrentWorkspace().getRoot().getLocation().toOSString());
		pthTest = (Path)pthTest.append(pth);
		File fileTest = pthTest.toFile();
		return (fileTest.exists()) && (fileTest.isFile());
	}
	
	public boolean resourceExists()
	{
		IPath pth = Path.fromPortableString(this.getPath());
		boolean bExists = checkExistsOnDisk(pth.toPortableString());
		if(!bExists) 
		{
			IPath ptsh = Path.fromPortableString(this.getPath());
			String strCommonNameCandidate = ptsh.segment(0);
			if(strCommonNameCandidate != null)
			{
				Collection<Project> prjs = ProjectConfig.getConfig().getProjects();
				for (Project project : prjs)
				{
					if(project.getCommonName().equalsIgnoreCase(strCommonNameCandidate))
					{
						IPath p = Path.fromOSString(project.getLocalMapping());
						String strLocalPrjNameCandidate = p.segment(p.segmentCount()-1);
						IProject prj = GuiUtil.getCurrentWorkspace().getRoot().getProject(strLocalPrjNameCandidate);
						if(prj != null)
						{
							pth = prj.getFullPath().append(pth.removeFirstSegments(1));
							bExists = checkExistsOnDisk(pth.toPortableString());
							if(bExists) this.res = GuiUtil.getCurrentWorkspace().getRoot().getFile(pth);
							break;
						}
					}
				}
			}
		}
		else
		{
			this.res = GuiUtil.getCurrentWorkspace().getRoot().getFile(pth);
		}
		return bExists;
	}
	
	public static EclipseLink createFromJavaElement(IJavaElement elem) throws URISyntaxException, UnsupportedJavaElement
	{
		if ((elem instanceof IMethod) || (elem instanceof IField))
		{
			String strPath = "";
			strPath = elem.getPath().toString();
			ArrayList<LinkParameter> params = new ArrayList<LinkParameter>();
			if (elem instanceof IMethod)
			{
				params.add(new LinkParameter(IEclipseLinkParamsTypes.METHOD_NAME, elem.getElementName()));
			}
			else if (elem instanceof IField)
			{
				params.add(new LinkParameter(IEclipseLinkParamsTypes.FIELD_NAME, elem.getElementName()));
			}
			Project prj = ProjectConfig.getConfig().resolveProjectFromIProject(elem.getJavaProject().getProject());
			return new EclipseLink(strPath, params,prj);
		}
		else if (elem instanceof ICompilationUnit)
		{
			ICompilationUnit cu = (ICompilationUnit)elem;
			try
			{
				Project prj = ProjectConfig.getConfig().resolveProjectFromIProject(cu.getJavaProject().getProject());
				return GuiUtil.createLinkEilertFromFile((IFile)cu.getCorrespondingResource(),prj);
			}
			catch(JavaModelException ex)
			{
				GuiUtil.handleException(ex);
				return null;
			}
		}
		else
		{
			throw new UnsupportedJavaElement(elem);
		}
	}
	
	public static EclipseLink createFromLineNumber(int nLine, IFile file)
	{
		if (nLine > -1)
		{
			ArrayList<LinkParameter> params = new ArrayList<LinkParameter>();
			params.add(new LinkParameter(IEclipseLinkParamsTypes.LINE_NUMBER, nLine + ""));
			EclipseLink eclLink = null;
			try
			{
				Project prj = ProjectConfig.getConfig().resolveProjectFromIProject(file.getProject());
				eclLink = new EclipseLink(file.getFullPath().toString(), params,prj);
			}
			catch(URISyntaxException ex)
			{
				ex.printStackTrace();
			}
			return eclLink;
		}
		else return null;
	}
}
